package org.apidesign.apifest08.currency;

import java.io.Serializable;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Currency;

/**
 * Default {@link Convertor} implementation. Exchange rate is stored as equivalents. It means if we have USD to CZK convertor and USD1 = CZK17 
 * we store 1 in sourceEquivalent and 17 in destinationEquivalent. This class is immutable.
 * @author lukas
 *
 */
class DefaultConvertor implements Convertor, Serializable {

	private static final long serialVersionUID = -1754789142402148099L;

	/**
	 * Equivalent in source currency.
	 */
	private final BigDecimal sourceEquivalent;
	
	/**
	 * Equivalent in destination currency.
	 */
	private final BigDecimal destinationEquivalent;
	
	private final Currency sourceCurrency;
	
	private final Currency destinationCurrency;
	
	public DefaultConvertor(BigDecimal sourceEquivalent, BigDecimal destinationEquivalent, Currency sourceCurrency, Currency destinationCurrency) {
		super();
		if (BigDecimal.ZERO.compareTo(sourceEquivalent)==0)
		{
			throw new IllegalArgumentException("Source equivalent amount can not be 0.");
		}
		if (BigDecimal.ZERO.compareTo(destinationEquivalent)==0)
		{
			throw new IllegalArgumentException("Destination equivalent amount can not be 0.");
		}
		this.sourceEquivalent = sourceEquivalent;
		this.destinationEquivalent = destinationEquivalent;
		this.sourceCurrency = sourceCurrency;
		this.destinationCurrency = destinationCurrency;
	}
	
	public Money convert(Money amount, Currency destinationCurrency) {
		if (amount==null)
		{
			throw new NullPointerException("Money is null");
		}
		if (destinationCurrency==null)
		{
			throw new NullPointerException("destionationCurrency is null");
		}
		if (isConversionInOpositeDirection(amount, destinationCurrency))
		{
			return revert().convert(amount, destinationCurrency);
		}
		if (!amount.getCurrency().equals(getSourceCurrency()))
		{
			throw new IllegalArgumentException("Can not convert from "+amount.getCurrency()+". Converts between "+getSourceCurrency()+" and "+getDestinationCurrency());
		}
		if (!getDestinationCurrency().equals(destinationCurrency))
		{
			throw new IllegalArgumentException("Can not convert to "+destinationCurrency+". Converts between "+getSourceCurrency()+" and "+getDestinationCurrency());
		}
		BigDecimal sourceAmount = amount.getAmount();
		BigDecimal destinationAmount = sourceAmount.multiply(destinationEquivalent).divide(sourceEquivalent, 2, RoundingMode.HALF_UP);
		return new MoneyImpl(destinationAmount, getDestinationCurrency());
	}

	/**
	 * Returns true, if the conversion is in oposit direction.
	 * @param amount
	 * @param destinationCurrency
	 * @return
	 */
	private boolean isConversionInOpositeDirection(Money amount,	Currency destinationCurrency) {
		return amount.getCurrency().equals(getDestinationCurrency()) && destinationCurrency.equals(getSourceCurrency());
	}


	public Convertor revert() {
		return new DefaultConvertor(destinationEquivalent, sourceEquivalent, destinationCurrency, sourceCurrency);
	}

	public BigDecimal getSourceEquivalent() {
		return sourceEquivalent;
	}

	public BigDecimal getDestinationEquivalent() {
		return destinationEquivalent;
	}

	public Currency getSourceCurrency() {
		return sourceCurrency;
	}

	public Currency getDestinationCurrency() {
		return destinationCurrency;
	}
	
	@Override
	public String toString() {
		return getClass().getName()+" converts "+getSourceCurrency()+" to "+getDestinationCurrency()+" "
			+getSourceCurrency()+getSourceEquivalent()+"="+getDestinationCurrency()+getDestinationEquivalent();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime
				* result
				+ ((destinationCurrency == null) ? 0 : destinationCurrency
						.hashCode());
		result = prime
				* result
				+ ((destinationEquivalent == null) ? 0 : destinationEquivalent
						.hashCode());
		result = prime * result
				+ ((sourceCurrency == null) ? 0 : sourceCurrency.hashCode());
		result = prime
				* result
				+ ((sourceEquivalent == null) ? 0 : sourceEquivalent.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof DefaultConvertor))
			return false;
		DefaultConvertor other = (DefaultConvertor) obj;
		if (destinationCurrency == null) {
			if (other.destinationCurrency != null)
				return false;
		} else if (!destinationCurrency.equals(other.destinationCurrency))
			return false;
		if (destinationEquivalent == null) {
			if (other.destinationEquivalent != null)
				return false;
		} else if (!destinationEquivalent.equals(other.destinationEquivalent))
			return false;
		if (sourceCurrency == null) {
			if (other.sourceCurrency != null)
				return false;
		} else if (!sourceCurrency.equals(other.sourceCurrency))
			return false;
		if (sourceEquivalent == null) {
			if (other.sourceEquivalent != null)
				return false;
		} else if (!sourceEquivalent.equals(other.sourceEquivalent))
			return false;
		return true;
	}


}
